@@ -1,11 +1,10 @@
 package Minion::Backend::Mango;
 use Mojo::Base 'Minion::Backend';
 
-our $VERSION = '0.97';
+our $VERSION = '0.98';
 
 use Mango;
 use Mango::BSON qw(bson_oid bson_time bson_doc);
-use Scalar::Util 'weaken';
 use Sys::Hostname 'hostname';
 
 has 'mango';
@@ -15,17 +14,14 @@ has prefix        => 'minion';
 has workers       => sub { $_[0]->mango->db->collection($_[0]->prefix . '.workers') };
 
 sub dequeue {
-  my ($self, $oid) = @_;
+  my ($self, $oid, $wait, $options) = @_;
 
-  # Capped collection for notifications
   $self->_notifications;
 
-  # Await notifications
-  $self->_await;
-  my $job = $self->_try($oid);
+  if (my $job = $self->_try($oid, $options)) { return $self->_job_info($job) }
 
-  return undef unless $self->_job_info($job);
-  return {args => $job->{args}, id => $job->{_id}, task => $job->{task}};
+  $self->_await;
+  return $self->_job_info($self->_try($oid, $options));
 }
 
 sub enqueue {
@@ -37,15 +33,17 @@ sub enqueue {
   $self->_notifications;
 
   my $doc = {
-    args    => $args,
-    created => bson_time,
-    delayed => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1),
+    args     => $args,
+    attempts => $options->{attempts} // 1,
+    created  => bson_time,
+    delayed  => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1),
     priority => $options->{priority} // 0,
+    queue    => $options->{queue}    // 'default',
+    retries  => 0,
     state    => 'inactive',
     task     => $task
   };
 
-  # Blocking
   my $oid = $self->jobs->insert($doc);
   $self->notifications->insert({c => 'created'});
   return $oid;
@@ -85,7 +83,7 @@ sub register_worker {
     && $self->workers->find_and_modify(
     {query => {_id => $id}, update => {'$set' => {notified => bson_time}}});
 
-  $self->jobs->ensure_index(bson_doc(state => 1, delayed => 1, task => 1));
+  $self->jobs->ensure_index(bson_doc(state => 1, delayed => 1, task => 1, queue => 1));
   $self->workers->insert({host => hostname, pid => $$, started => bson_time, notified => bson_time});
 }
 
@@ -119,18 +117,20 @@ sub repair {
 sub reset { $_->options && $_->drop for $_[0]->workers, $_[0]->jobs }
 
 sub retry_job {
-  my ($self, $oid) = (shift, shift);
+  my ($self, $oid, $retries) = (shift, shift, shift);
   my $options = shift // {};
 
-  my $query = {_id => $oid, state => {'$in' => [qw(failed finished)]}};
+  my $query = {_id => $oid, retries => $retries, state => {'$in' => [qw(failed finished)]}};
   my $update = {
     '$inc' => {retries => 1},
     '$set' => {
       retried => bson_time,
       state   => 'inactive',
-      delayed => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1)
+      delayed => bson_time($options->{delay} ? (time + $options->{delay}) * 1000 : 1),
+      (defined $options->{priority} ? (priority => $options->{priority}) : ()),
+      (defined $options->{queue}    ? (queue    => $options->{queue})    : ())
     },
-    '$unset' => {map { $_ => '' } qw(finished result started worker)}
+    '$unset' => {map { $_ => '' } qw(finished started worker)}
   };
 
   return !!$self->jobs->update($query, $update)->{n};
@@ -169,11 +169,13 @@ sub _job_info {
   return undef unless my $job = shift;
   return {
     args     => $job->{args},
+    attempts => $job->{attempts},
     created  => $job->{created} ? $job->{created}->to_epoch : undef,
     delayed  => $job->{delayed} ? $job->{delayed}->to_epoch : undef,
     finished => $job->{finished} ? $job->{finished}->to_epoch : undef,
     id       => $job->{_id},
     priority => $job->{priority},
+    queue    => $job->{queue},
     result   => $job->{result},
     retried  => $job->{retried} ? $job->{retried}->to_epoch : undef,
     retries => $job->{retries} // 0,
@@ -195,16 +197,17 @@ sub _notifications {
 }
 
 sub _try {
-  my ($self, $oid) = @_;
+  my ($self, $oid, $options) = @_;
 
   my $doc = {
     query => bson_doc(
-      delayed => {'$lt' => bson_time},
       state   => 'inactive',
-      task    => {'$in' => [keys %{$self->minion->tasks}]}
+      delayed => {'$lt' => bson_time},
+      task    => {'$in' => [keys %{$self->minion->tasks}]},
+      queue   => {'$in' => $options->{queues} || ['default']}
     ),
-    fields => {args     => 1, task => 1},
-    sort   => {priority => -1},
+    fields => bson_doc(args     => 1,  attempts => 1, retries => 1, task => 1),
+    sort   => bson_doc(priority => -1, created  => 1),
     update => {'$set' => {started => bson_time, state => 'active', worker => $oid}},
     new    => 1
   };
@@ -213,11 +216,10 @@ sub _try {
 }
 
 sub _update {
-  my ($self, $fail, $oid, $err) = @_;
+  my ($self, $fail, $oid, $retries, $result) = @_;
 
-  my $update = {finished => bson_time, state => $fail ? 'failed' : 'finished'};
-  $update->{result} = $err if $fail;
-  my $query = {_id => $oid, state => 'active'};
+  my $update = {finished => bson_time, result => $result, state => $fail ? 'failed' : 'finished'};
+  my $query = {_id => $oid, retries => $retries, state => 'active'};
   return !!$self->jobs->update($query, {'$set' => $update})->{n};
 }
 
@@ -300,11 +302,24 @@ L<Minion::Backend::Mango> inherits all methods from L<Minion::Backend> and imple
 
 =head2 dequeue
 
-  my $info = $backend->dequeue($worker_id, 0.5);
+  my $job_info = $backend->dequeue($worker_id, 0.5);
+  my $job_info = $backend->dequeue($worker_id, 0.5, {queues => ['default']});
 
 Wait for job, dequeue it and transition from C<inactive> to C<active> state or
 return C<undef> if queue was empty.
 
+These options are currently available:
+
+=over 2
+
+=item queues
+
+  queues => ['important']
+
+One or more queues to dequeue jobs from, defaults to C<default>.
+
+=back
+
 =head2 enqueue
 
   my $job_id = $backend->enqueue('foo');
@@ -315,6 +330,12 @@ Enqueue a new job with C<inactive> state. These options are currently available:
 
 =over 2
 
+=item attempts
+
+  attempts => 25
+
+Number of times performing this job will be attempted, defaults to C<1>.
+
 =item delay
 
   delay => 10
@@ -327,6 +348,12 @@ Delay job for this many seconds from now.
 
 Job priority, defaults to C<0>.
 
+=item queue
+
+  queue => 'important'
+
+Queue to put job in, defaults to C<default>.
+
 =back
 
 =head2 fail_job
@@ -427,6 +454,18 @@ These options are currently available:
 
 Delay job for this many seconds (from now).
 
+=item priority
+
+  priority => 5
+
+Job priority.
+
+=item queue
+
+  queue => 'important'
+
+Queue to put job in.
+
 =back
 
 =head2 stats
